/* TroiaPreview
 * version 0.4 date 2004-10-23
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __APPLE__
# include "/System/Library/Frameworks/OpenGL.framework/Versions/A/Headers/gl.h"
# include "/System/Library/Frameworks/OpenGL.framework/Versions/A/Headers/glu.h"
#else
# include <GL/gl.h>
# include <GL/glu.h>
#endif

#include "Vector.h"
#include "Globals.h"
#include "Pixel.h"
#include "Frame.h"

//static frame
unsigned char FrameWall1[PixelZ][PixelX];
unsigned char FrameWall2[PixelZ][PixelY];
unsigned char FrameWall3[PixelZ][PixelX];
unsigned char FrameWall4[PixelZ][PixelY];
unsigned char FrameCeiling[PixelY][PixelX];
unsigned char FrameFloor[PixelY][PixelX];

//temporary frame
unsigned char FrameTmpWall1[PixelZ][PixelX];
unsigned char FrameTmpWall2[PixelZ][PixelY];
unsigned char FrameTmpWall3[PixelZ][PixelX];
unsigned char FrameTmpWall4[PixelZ][PixelY];
unsigned char FrameTmpCeiling[PixelY][PixelX];
unsigned char FrameTmpFloor[PixelY][PixelX];

//initialize frame with constant color
void FrameSet( unsigned char Color )
{
  int X, Y, Z;

  for( Z = 0; Z < PixelZ; Z++ )
    for( X = 0; X < PixelX; X++ )
      FrameWall1[Z][X] = Color;
  for( Z = 0; Z < PixelZ; Z++ )
    for( Y = 0; Y < PixelY; Y++ )
      FrameWall2[Z][Y] = Color;
  for( Z = 0; Z < PixelZ; Z++ )
    for( X = 0; X < PixelX; X++ )
      FrameWall3[Z][X] = Color;
  for( Z = 0; Z < PixelZ; Z++ )
    for( Y = 0; Y < PixelY; Y++ )
      FrameWall4[Z][Y] = Color;
  for( Y = 0; Y < PixelY; Y++ )
    for( X = 0; X < PixelX; X++ )
      FrameCeiling[Y][X] = Color;
  for( Y = 0; Y < PixelY; Y++ )
    for( X = 0; X < PixelX; X++ )
      FrameFloor[Y][X] = Color;
}

//initialize frame with random data
void FrameRandomize( unsigned char ColorMin, unsigned char ColorMax )
{
  unsigned char Color;
  int Range, X, Y, Z;

  //correct minimum and maximum color
  if( ColorMin > ColorMax )
  {
    Color = ColorMin;
    ColorMin = ColorMax;
    ColorMax = Color;
  }
  //calculate color range
  Range = ColorMax - ColorMin;

  for( Z = 0; Z < PixelZ; Z++ )
    for( X = 0; X < PixelX; X++ )
      FrameWall1[Z][X] = (unsigned char)(rand( ) % Range) + ColorMin;
  for( Z = 0; Z < PixelZ; Z++ )
    for( Y = 0; Y < PixelY; Y++ )
      FrameWall2[Z][Y] = (unsigned char)(rand( ) % Range) + ColorMin;
  for( Z = 0; Z < PixelZ; Z++ )
    for( X = 0; X < PixelX; X++ )
      FrameWall3[Z][X] = (unsigned char)(rand( ) % Range) + ColorMin;
  for( Z = 0; Z < PixelZ; Z++ )
    for( Y = 0; Y < PixelY; Y++ )
      FrameWall4[Z][Y] = (unsigned char)(rand( ) % Range) + ColorMin;
  for( Y = 0; Y < PixelY; Y++ )
    for( X = 0; X < PixelX; X++ )
      FrameCeiling[Y][X] = (unsigned char)(rand( ) % Range) + ColorMin;
  for( Y = 0; Y < PixelY; Y++ )
    for( X = 0; X < PixelX; X++ )
      FrameFloor[Y][X] = (unsigned char)(rand( ) % Range) + ColorMin;
}

//get a pixel of frame
unsigned char FrameGetPixel( tPixel3D Pix )
{
  switch( Pix.Wall )
  {
    case 1:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        return FrameWall1[Pix.Y][Pix.X];
      break;
    case 2:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        return FrameWall2[Pix.Y][Pix.X];
      break;
    case 3:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        return FrameWall3[Pix.Y][Pix.X];
      break;
    case 4:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        return FrameWall4[Pix.Y][Pix.X];
      break;
    case 5:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        return FrameCeiling[Pix.Y][Pix.X];
      break;
    case 6:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        return FrameFloor[Pix.Y][Pix.X];
      break;
  }
  return 0;
}

//set a pixel of frame
void FrameSetPixel( tPixel3D Pix, unsigned char Color )
{
  switch( Pix.Wall )
  {
    case 1:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        FrameWall1[Pix.Y][Pix.X] = Color;
      break;
    case 2:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        FrameWall2[Pix.Y][Pix.X] = Color;
      break;
    case 3:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        FrameWall3[Pix.Y][Pix.X] = Color;
      break;
    case 4:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        FrameWall4[Pix.Y][Pix.X] = Color;
      break;
    case 5:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        FrameCeiling[Pix.Y][Pix.X] = Color;
      break;
    case 6:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        FrameFloor[Pix.Y][Pix.X] = Color;
      break;
  }
}

//draw a line
void FrameLine( tPixel3D Start, unsigned char Dir, unsigned int Len, unsigned char Color )
{
  for( ; Len > 0; Len-- )
  {
    FrameSetPixel( Start, Color );
    Pixel3DMove( Start, Dir, &Start, &Dir );
  }
}

//draw a filled rectangle
void FrameRect( tPixel3D Start, unsigned char Dir, unsigned int Width, unsigned int Height, unsigned char Color )
{
  tPixel3D Row, Pix;
  unsigned char Down, Right;
  unsigned int X, Y;

  Row = Start;
  Down = DirRotate( Dir, 2 );
  for( Y = 0; Y < Height; Y++ )
  {
    Pix = Row;
    Right = DirRotate( Down, -2 );
    for( X = 0; X < Width; X++ )
    {
      FrameSetPixel( Pix, Color );
      Pixel3DMove( Pix, Right, &Pix, &Right );
    }
    Pixel3DMove( Row, Down, &Row, &Down );
  }
}

//draw an image
void FrameImage( tPixel3D Start, unsigned char Dir, unsigned int Width, unsigned int Height, unsigned char * pData )
{
  tPixel3D Row, Pix;
  unsigned char Down, Right, Color;
  unsigned int X, Y, I;

  Row = Start;
  Down = DirRotate( Dir, 2 );
  I = 0;
  for( Y = 0; Y < Height; Y++ )
  {
    Pix = Row;
    Right = DirRotate( Down, -2 );
    for( X = 0; X < Width; X++ )
    {
      Color = (unsigned char)( ((unsigned short)pData[I+0]
                  + (unsigned short)pData[I+1]
                  + (unsigned short)pData[I+2]) / 3 ); //color is avarage from R, G, B
      FrameSetPixel( Pix, Color );
      Pixel3DMove( Pix, Right, &Pix, &Right );
      I += 3; //pixels are 3 byte (R, G, B)
    }
    Pixel3DMove( Row, Down, &Row, &Down );
    I = (I + 3) & ~3; //lines are 4-byte aligned
  }
}

//get new color for frame pixels by callback function
void FrameColorByCallback( unsigned char (*Callback) ( tPixel3D Pixel, void * pContext ), void * pContext )
{
  tPixel3D Pixel;

  //call callback for every pixel
  Pixel.Wall = 1;
  for( Pixel.Y = 0; Pixel.Y < PixelZ; Pixel.Y++ )
    for( Pixel.X = 0; Pixel.X < PixelX; Pixel.X++ )
      FrameWall1[Pixel.Y][Pixel.X] = Callback( Pixel, pContext );
  Pixel.Wall = 2;
  for( Pixel.Y = 0; Pixel.Y < PixelZ; Pixel.Y++ )
    for( Pixel.X = 0; Pixel.X < PixelY; Pixel.X++ )
      FrameWall2[Pixel.Y][Pixel.X] = Callback( Pixel, pContext );
  Pixel.Wall = 3;
  for( Pixel.Y = 0; Pixel.Y < PixelZ; Pixel.Y++ )
    for( Pixel.X = 0; Pixel.X < PixelX; Pixel.X++ )
      FrameWall3[Pixel.Y][Pixel.X] = Callback( Pixel, pContext );
  Pixel.Wall = 4;
  for( Pixel.Y = 0; Pixel.Y < PixelZ; Pixel.Y++ )
    for( Pixel.X = 0; Pixel.X < PixelY; Pixel.X++ )
      FrameWall4[Pixel.Y][Pixel.X] = Callback( Pixel, pContext );
  Pixel.Wall = 5;
  for( Pixel.Y = 0; Pixel.Y < PixelY; Pixel.Y++ )
    for( Pixel.X = 0; Pixel.X < PixelX; Pixel.X++ )
      FrameCeiling[Pixel.Y][Pixel.X] = Callback( Pixel, pContext );
  Pixel.Wall = 6;
  for( Pixel.Y = 0; Pixel.Y < PixelY; Pixel.Y++ )
    for( Pixel.X = 0; Pixel.X < PixelX; Pixel.X++ )
      FrameFloor[Pixel.Y][Pixel.X] = Callback( Pixel, pContext );
}

//set frame to 2D data
//p2D must point to a buffer with an entire frame in 2D format
void FrameSet2D( unsigned char * p2D )
{
  int X, Y, Z, I, J;

  for( Z = 0, I = PixelY / 2 * (2 * PixelX + 2 * PixelY); Z < PixelZ; Z++ )
  {
    for( X = 0; X < PixelX; X++, I++ )
      FrameWall1[Z][X] = p2D[I];
    for( Y = 0; Y < PixelY; Y++, I++ )
      FrameWall2[Z][Y] = p2D[I];
    for( X = 0; X < PixelX; X++, I++ )
      FrameWall3[Z][X] = p2D[I];
    for( Y = 0; Y < PixelY; Y++, I++ )
      FrameWall4[Z][Y] = p2D[I];
  }
  for( Y = PixelY / 2 - 1, I = 0; Y >= 0; Y--, I += 2 * PixelX + 2 * PixelY )
    for( X = 0, J = I; X < PixelX; X++, J++ )
      FrameCeiling[Y][X] = p2D[J];
  for( Y = PixelY / 2, I = PixelX + PixelY; Y < PixelY; Y++, I += 2 * PixelX + 2 * PixelY )
    for( X = PixelX - 1, J = I; X >= 0; X--, J++ )
      FrameCeiling[Y][X] = p2D[J];
  for( Y = 0, I = 0; Y < PixelY / 2; Y++, I += 2 * PixelX + 2 * PixelY )
    for( X = 0, J = I; X < PixelX; X++, J++ )
      FrameFloor[Y][X] = p2D[J];
  for( Y = PixelY - 1, I = PixelX + PixelY; Y >= PixelY / 2; Y--, I += 2 * PixelX + 2 * PixelY )
    for( X = PixelX - 1, J = I; X >= 0; X--, J++ )
      FrameFloor[Y][X] = p2D[J];
}

//copy static frame into temporary frame
void FrameCopyTmp( )
{
  memcpy( FrameTmpWall1, FrameWall1, sizeof( FrameWall1 ) );
  memcpy( FrameTmpWall2, FrameWall2, sizeof( FrameWall2 ) );
  memcpy( FrameTmpWall3, FrameWall3, sizeof( FrameWall3 ) );
  memcpy( FrameTmpWall4, FrameWall4, sizeof( FrameWall4 ) );
  memcpy( FrameTmpCeiling, FrameCeiling, sizeof( FrameCeiling ) );
  memcpy( FrameTmpFloor, FrameFloor, sizeof( FrameFloor ) );
}

//get a pixel of temporary frame
unsigned char FrameTmpGetPixel( tPixel3D Pix )
{
  switch( Pix.Wall )
  {
    case 1:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        return FrameTmpWall1[Pix.Y][Pix.X];
      break;
    case 2:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        return FrameTmpWall2[Pix.Y][Pix.X];
      break;
    case 3:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        return FrameTmpWall3[Pix.Y][Pix.X];
      break;
    case 4:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        return FrameTmpWall4[Pix.Y][Pix.X];
      break;
    case 5:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        return FrameTmpCeiling[Pix.Y][Pix.X];
      break;
    case 6:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        return FrameTmpFloor[Pix.Y][Pix.X];
      break;
  }
  return 0;
}

//set a pixel of temporary frame
void FrameTmpSetPixel( tPixel3D Pix, unsigned char Color )
{
  switch( Pix.Wall )
  {
    case 1:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        FrameTmpWall1[Pix.Y][Pix.X] = Color;
      break;
    case 2:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        FrameTmpWall2[Pix.Y][Pix.X] = Color;
      break;
    case 3:
      if( Pix.X < PixelX && Pix.Y < PixelZ )
        FrameTmpWall3[Pix.Y][Pix.X] = Color;
      break;
    case 4:
      if( Pix.X < PixelY && Pix.Y < PixelZ )
        FrameTmpWall4[Pix.Y][Pix.X] = Color;
      break;
    case 5:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        FrameTmpCeiling[Pix.Y][Pix.X] = Color;
      break;
    case 6:
      if( Pix.X < PixelX && Pix.Y < PixelY )
        FrameTmpFloor[Pix.Y][Pix.X] = Color;
      break;
  }
}

//show a pixel
//Pos: top left corner of pixel
//Right: vector from left to right side
//Bottom: vector from top to bottom side
//Color: greyscale value of pixel (0..255)
static void ShowPixel( tVector Pos, tVector Right, tVector Bottom, unsigned char Color )
{
  float fColor;

  //set Color
  fColor = (float)Color * (1.0f / 255.0f);
  glColor3f( fColor, fColor, fColor );

  //draw a quad
  glBegin( GL_QUADS );
  glTexCoord2d( 0.0, 0.0 ); //top left
  glVertex3d( Pos.X, Pos.Y, Pos.Z );
  Add( &Pos, Pos, Right ); //top right
  glTexCoord2d( 1.0, 0.0 );
  glVertex3d( Pos.X, Pos.Y, Pos.Z );
  Add( &Pos, Pos, Bottom ); //bottom right
  glTexCoord2d( 1.0, 1.0 );
  glVertex3d( Pos.X, Pos.Y, Pos.Z );
  Sub( &Pos, Pos, Right ); //bottom left
  glTexCoord2d( 0.0, 1.0 );
  glVertex3d( Pos.X, Pos.Y, Pos.Z );
  glEnd( );
}

//show pixels
//displays all pixels in the temoprary frame using OpenGL
void ShowPixels( )
{
  //local variables
  long X, Y, Z;
  tVector Pos, Right, Bottom;

  //show wall1
  Pos.Y = PixelY * PixelSize;
  SetVector( &Right, PixelSize, 0.0, 0.0 );
  SetVector( &Bottom, 0.0, 0.0, -PixelSize );
  for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
    for( X = 0, Pos.X = 0.0; X < PixelX; X++, Pos.X += PixelSize )
      ShowPixel( Pos, Right, Bottom, FrameTmpWall1[Z][X] );
  //show wall2
  Pos.X = PixelX * PixelSize;
  SetVector( &Right, 0.0, -PixelSize, 0.0 );
  SetVector( &Bottom, 0.0, 0.0, -PixelSize );
  for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
    for( Y = 0, Pos.Y = PixelY * PixelSize; Y < PixelY; Y++, Pos.Y -= PixelSize )
      ShowPixel( Pos, Right, Bottom, FrameTmpWall2[Z][Y] );
  //show wall3
  Pos.Y = 0.0;
  SetVector( &Right, -PixelSize, 0.0, 0.0 );
  SetVector( &Bottom, 0.0, 0.0, -PixelSize );
  for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
    for( X = 0, Pos.X = PixelX * PixelSize; X < PixelX; X++, Pos.X -= PixelSize )
      ShowPixel( Pos, Right, Bottom, FrameTmpWall3[Z][X] );
  //show wall4
  Pos.X = 0.0;
  SetVector( &Right, 0.0, PixelSize, 0.0 );
  SetVector( &Bottom, 0.0, 0.0, -PixelSize );
  for( Z = 0, Pos.Z = PixelZ * PixelSize; Z < PixelZ; Z++, Pos.Z -= PixelSize )
    for( Y = 0, Pos.Y = 0.0; Y < PixelY; Y++, Pos.Y += PixelSize )
      ShowPixel( Pos, Right, Bottom, FrameTmpWall4[Z][Y] );
  //show ceiling
  Pos.Z = PixelZ * PixelSize;
  SetVector( &Right, PixelSize, 0.0, 0.0 );
  SetVector( &Bottom, 0.0, PixelSize, 0.0 );
  for( Y = 0, Pos.Y = (PixelY - 1) * PixelSize; Y < PixelY; Y++, Pos.Y -= PixelSize )
    for( X = 0, Pos.X = 0.0; X < PixelX; X++, Pos.X += PixelSize )
      ShowPixel( Pos, Right, Bottom, FrameTmpCeiling[Y][X] );
  //show floor
  Pos.Z = 0.0;
  SetVector( &Right, PixelSize, 0.0, 0.0 );
  SetVector( &Bottom, 0.0, -PixelSize, 0.0 );
  for( Y = 0, Pos.Y = PixelY * PixelSize; Y < PixelY; Y++, Pos.Y -= PixelSize )
    for( X = 0, Pos.X = 0.0; X < PixelX; X++, Pos.X += PixelSize )
      ShowPixel( Pos, Right, Bottom, FrameTmpFloor[Y][X] );
}
