/* TroiaPreview
 * version 0.4 date 2004-10-23
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <string.h>

#include "Font.h"
#include "Pixel.h"
#include "Frame.h"
#include "Text.h"

//get width of text in pixels
unsigned int TextWidth( char * pTxt )
{
  return strlen( pTxt ) * FontSizeX;
}

//output text
//top left corner of text is at pixel Pix
//write in direction Dir (encoded as the number on numpad (e.g. 8 for up))
//if a color is not valid (0..255), it will not be drawn
//StaticFrame decides if text is written to static or temporary frame
void TextOutput( char * pTxt, tPixel3D TopLeft, unsigned char Dir,
         unsigned short ForeColor, unsigned short BackColor,
         int StaticFrame )
{
  tPixel3D Pix;
  unsigned char Down;
  unsigned char CharFontData[FontSizeY];
  int X, Y;
  unsigned short Color;

  //output all characters
  for( ; *pTxt != 0; pTxt++ )
  {
    //get pointer to font data for this character
    for( Y = 0; Y < FontSizeY; Y++ )
      CharFontData[Y] = FontData[(unsigned char)*pTxt][Y];
    //output all columns of character
    for( X = FontSizeX - 1; X >= 0; X-- )
    {
      //start downwards at top of row
      Pix = TopLeft;
      Down = DirRotate( Dir, 2 );
      //output all rows of character
      for( Y = 0; Y < FontSizeY; Y++ )
      {
        //set pixel
        Color = CharFontData[Y] & (1 << X) ? ForeColor : BackColor;
        if( Color < 256 )
        {
          if( StaticFrame )
            FrameSetPixel( Pix, (unsigned char)Color );
          else
            FrameTmpSetPixel( Pix, (unsigned char)Color );
        }
        //go one pixel down
        Pixel3DMove( Pix, Down, &Pix, &Down );
      }
      //go one pixel right
      Pixel3DMove( TopLeft, Dir, &TopLeft, &Dir );
    }
  } //for( ; *pTxt ...
}

//output text - extended version
//top left corner of text is at pixel Pix
//write in direction Dir (encoded as the number on numpad (e.g. 8 for up))
//if a color is not valid (0..255), it will not be drawn
//StaticFrame decides if text is written to static or temporary frame
//draw <Pad> empty pixels before the beginning of the text
//skip <Skip> pixels of the beginning of the text
//draw <Width> pixels in total (including padding at beginning and possible padding at end)
void TextOutputEx( char * pTxt, tPixel3D TopLeft, unsigned char Dir,
           unsigned short ForeColor, unsigned short BackColor,
           int StaticFrame, 
           unsigned int Pad, unsigned int Skip, unsigned int Width )
{
  tPixel3D Pix;
  unsigned char Down;
  unsigned char CharFontData[FontSizeY];
  int X, Y;
  unsigned short Color;

  //draw padding at beginning
  for( ; Pad > 0 && Width > 0; Pad--, Width-- )
  {
    //start downwards at top of row
    Pix = TopLeft;
    Down = DirRotate( Dir, 2 );
    //if not transparent
    if( BackColor < 256 )
    {
      //output all rows
      for( Y = 0; Y < FontSizeY; Y++ )
      {
        //set pixel
        if( StaticFrame )
          FrameSetPixel( Pix, (unsigned char)BackColor );
        else
          FrameTmpSetPixel( Pix, (unsigned char)BackColor );
        //go one pixel down
        Pixel3DMove( Pix, Down, &Pix, &Down );
      }
    }
    //go one pixel right
    Pixel3DMove( TopLeft, Dir, &TopLeft, &Dir );
  }

  //skip entire characters at beginning of text
  for( ; Skip >= FontSizeX && *pTxt != 0; Skip -=FontSizeX, pTxt++ );

  //output all characters
  for( ; *pTxt != 0 && (Skip > 0 || Width > 0); pTxt++ )
  {
    //get pointer to font data for this character
    for( Y = 0; Y < FontSizeY; Y++ )
      CharFontData[Y] = FontData[(unsigned char)*pTxt][Y];
    //output all columns of character
    for( X = FontSizeX - 1; X >= 0; X-- )
    {
      //still skipping pixels at beginning of text
      if( Skip > 0 )
        Skip--;
      //not skipping
      else
      {
        //start downwards at top of row
        Pix = TopLeft;
        Down = DirRotate( Dir, 2 );
        //output all rows
        for( Y = 0; Y < FontSizeY; Y++ )
        {
          //set pixel
          Color = CharFontData[Y] & (1 << X) ? ForeColor : BackColor;
          if( Color < 256 )
          {
            if( StaticFrame )
              FrameSetPixel( Pix, (unsigned char)Color );
            else
              FrameTmpSetPixel( Pix, (unsigned char)Color );
          }
          //go one pixel down
          Pixel3DMove( Pix, Down, &Pix, &Down );
        }
        //go one pixel right
        Pixel3DMove( TopLeft, Dir, &TopLeft, &Dir );
        //count down width to draw
        Width--;
        if( Width == 0 )
          return;
      }
    }
  } //for( ; *pTxt ...

  //if not transparent
  if( BackColor < 256 )
  {
    //draw padding at end
    for( ; Width > 0; Width-- )
    {
      //start downwards at top of row
      Pix = TopLeft;
      Down = DirRotate( Dir, 2 );
      //output all rows
      for( Y = 0; Y < FontSizeY; Y++ )
      {
        //set pixel
        if( StaticFrame )
          FrameSetPixel( Pix, (unsigned char)BackColor );
        else
          FrameTmpSetPixel( Pix, (unsigned char)BackColor );
        //go one pixel down
        Pixel3DMove( Pix, Down, &Pix, &Down );
      }
      //go one pixel right
      Pixel3DMove( TopLeft, Dir, &TopLeft, &Dir );
    }
  }
}
