/* TroiaPreview
 * version 0.4 date 2004-10-23
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 *
 * parameters:
 *   script=<file>:                                 load script file
 *   log=<file>:                                    generate log file
 *   window=<width>x<height>:                       use window mode
 *   fullscreen=<width>x<height>:<bits>@<freqency>: use fullscreen mode
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __APPLE__
# include "/System/Library/Frameworks/OpenGL.framework/Versions/A/Headers/gl.h"
# include "/System/Library/Frameworks/OpenGL.framework/Versions/A/Headers/glu.h"
# include "/System/Library/Frameworks/GLUT.framework/Versions/A/Headers/glut.h"
#else
# include <GL/gl.h>
# include <GL/glu.h>
# include <GL/glut.h>
#endif

#include "Tools.h"
#include "Vector.h"
#include "Font.h"
#include "Images.h"
#include "Log.h"
#include "Textures.h"
#include "Console.h"
#include "Globals.h"
#include "Frame.h"
#include "Animations.h"
#include "Life.h"
#include "Worms.h"
#include "StreamRecv.h"
#include "Ticker.h"
#include "Tracking.h"
#include "Commands.h"

//global variables
int FullScreen = 1;
long ScreenX; //display-resolution
long ScreenY;
long ScreenBits;
long ScreenFreq;
#define ScreenY_X ((float)ScreenY / (float)ScreenX)

//glut display callback
static void GlutDisplay( )
{
  //local variables
  long X, Y, I;
  float fX, fY, fDX, fDY;

  //clear picture and depth-information
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  //settings for background and console
  glMatrixMode( GL_PROJECTION ); //set projection
  glLoadIdentity( );
  glMatrixMode( GL_MODELVIEW ); //set view
  glLoadIdentity( );
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL ); //draw filled polygons
  glEnable( GL_TEXTURE_2D ); //textures on

  //no antialiasing
  glDisable( GL_POINT_SMOOTH );
  glDisable( GL_LINE_SMOOTH );
  glDisable( GL_POLYGON_SMOOTH );
  glDisable( GL_BLEND );

  //draw background
  if( Back )
  {
    glColor3f( 1.0f, 1.0f, 1.0f );
    fDX = 2.0f / (float)BackTexCntX;
    fDY = 2.0f / (float)BackTexCntY;
    for( Y = 0; Y < BackTexCntY; Y++ )
    for( X = 0; X < BackTexCntX; X++ )
    {
      I = Y * BackTexCntX + X;
      fX = -1.0f + X * fDX;
      fY = 1.0f - Y * fDY;
      glBindTexture( GL_TEXTURE_2D, BackTexFirstNo + I );
      glBegin( GL_QUADS );
      glNormal3f( 0.0f, 0.0f, 1.0f );
      glTexCoord2f( 0.0f, 0.0f );
      glVertex3f( fX, fY, 1.0f );
      glTexCoord2f( 1.0f, 0.0f );
      glVertex3f( fX + fDX, fY, 1.0f );
      glTexCoord2f( 1.0f, 1.0f );
      glVertex3f( fX + fDX, fY - fDY, 1.0f );
      glTexCoord2f( 0.0f, 1.0f );
      glVertex3f( fX, fY - fDY, 1.0f );
      glEnd( );
    }
  }

  //draw console
  if( Console )
  {
    fX = 512.0f / (float)ScreenX;
    fY = 512.0f / (float)ScreenY;
    glColor3f( 1.0f, 1.0f, 1.0f );
    glBindTexture( GL_TEXTURE_2D, ConsoleTexNo );
    glBegin( GL_QUADS );
    glNormal3f( 0.0f, 0.0f, 1.0f );
    glTexCoord2f( 0.0f, 0.0f );
    glVertex3f( -1.0f, -1.0f, -1.0f );
    glTexCoord2f( 1.0f, 0.0f );
    glVertex3f( -1.0f + fX, -1.0f, -1.0f );
    glTexCoord2f( 1.0f, 1.0f );
    glVertex3f( -1.0f + fX, -1.0f + fY, -1.0f );
    glTexCoord2f( 0.0f, 1.0f );
    glVertex3f( -1.0f, -1.0f + fY, -1.0f );
    glEnd( );
  }

  //settings for objects in space
  glMatrixMode( GL_PROJECTION ); //set projection
  glLoadIdentity( );
  glFrustum( -0.01, 0.01, -0.01 * ScreenY_X, 0.01 * ScreenY_X, 0.01, 20.0 );
  glMatrixMode( GL_MODELVIEW ); //set view
  glLoadIdentity( );
  gluLookAt( Pos.X, Pos.Y, Pos.Z, Pos.X + Front.X, Pos.Y + Front.Y, Pos.Z + Front.Z, 0.0, 0.0, 1.0 );

  //enable antialiasing ?
  if( AntiAlias )
  {
    glEnable( GL_POINT_SMOOTH );
    glEnable( GL_LINE_SMOOTH );
    glEnable( GL_POLYGON_SMOOTH );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    //glHint( GL_POLYGON_SMOOTH_HINT, GL_NICEST );
     }

  //bind pixel texture
  glBindTexture( GL_TEXTURE_2D, PixTexNo );
  //enable texture
  glEnable( GL_TEXTURE_2D );
  //set polygon mode
  glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
  //show pixels
  ShowPixels( );

  //show picture
  glutSwapBuffers( );
}

//formula for interpolation
//A=f(-1), B=f(0), C=f(1), D=f(2), f(X) wanted
#define Move_f(A,B,C,D) (-A/6.0 + B/2.0 - C/2.0 + D/6.0) * XXX + (A/2.0 - B + C/2.0) * XX + (-A/3.0 - B/2 + C - D/6.0)*X + B

//move cam
static void MoveCam( )
{
  //local variables
  double CurrentTime, CurPt, X, XX, XXX;
  long CurPtInt;
  tVector P1, P2, P3, P4;

  //cam is not moving
  if( !CamMoving )
    return;

  //get time
  CurrentTime = TimeGet( );

  //one cam-point
  if( CamptsCnt == 1 )
  {
    //move camera to that cam-point
    Pos = pCampts[0].Pos;
    Front = pCampts[0].Front;
  }
  //more cam-points
  else if( CamptsCnt > 1 )
  {
    //get current cam-point-number
    CurPt = DoubleModulo( CamptSpeed * (CurrentTime - CamptRefTime), CamptsCnt );
    //get points for interpolation
    CurPtInt = (long)CurPt;
    X = CurPt - (double)CurPtInt;
    XX = X * X;
    XXX = XX * X;
    //get points for pos-interpolation
    P1 = pCampts[(CurPtInt + CamptsCnt - 1) % CamptsCnt].Pos;
    P2 = pCampts[CurPtInt].Pos;
    P3 = pCampts[(CurPtInt + 1) % CamptsCnt].Pos;
    P4 = pCampts[(CurPtInt + 2) % CamptsCnt].Pos;
    //pos-interpolation
    //P1=f(-1), P2=f(0), P3=f(1), P4=f(2), f(X) wanted
    Pos.X = Move_f( P1.X, P2.X, P3.X, P4.X );
    Pos.Y = Move_f( P1.Y, P2.Y, P3.Y, P4.Y );
    Pos.Z = Move_f( P1.Z, P2.Z, P3.Z, P4.Z );
    //get points for front-interpolation
    P1 = pCampts[(CurPtInt + CamptsCnt - 1) % CamptsCnt].Front;
    P2 = pCampts[CurPtInt].Front;
    P3 = pCampts[(CurPtInt + 1) % CamptsCnt].Front;
    P4 = pCampts[(CurPtInt + 2) % CamptsCnt].Front;
    //front-interpolation
    //P1=f(-1), P2=f(0), P3=f(1), P4=f(2), f(X) wanted
    Front.X = Move_f( P1.X, P2.X, P3.X, P4.X );
    Front.Y = Move_f( P1.Y, P2.Y, P3.Y, P4.Y );
    Front.Z = Move_f( P1.Z, P2.Z, P3.Z, P4.Z );
  }

  //repair position and front direction
  RepairPosFront( );
}

//glut idle callback
static void GlutIdle( )
{
  //process a few script-commands
  ProcessScriptBufferCommands( );

  //calls tasks working on static frame
  AnimTask( );
  LifeTask( );
  WormTask( );
  StreamRecvTask( );

  //copy static frame into temporary frame
  FrameCopyTmp( );

  //call tasks working on temporary frame
  TickerTask( );
  TrackingTask( );

  //do automatic cam movement
  MoveCam( );

  //refresh console
  if( Console )
    ConsoleRefresh( );

  //schedule redisplay
  glutPostRedisplay( );

  //end program if requested
  if( RequestEnd )
    exit( 0 ); //this is the only way to exit glutMainLoop( )
}

//glut keyboard callback
static void GlutKeyboard ( unsigned char Key, int X, int Y )
{
  //local variables
  char Line[256];
  tVector Dir, Top, Left;

  //keep compiler happy
  X = 0;
  Y = 0;

  //get direction over ground
  Dir = Front;
  Dir.Z = 0;
  Normalize( &Dir );
  //get top and left vectors
  SetVector( &Top, 0.0, 0.0, 1.0 );
  Left = Dir;
  Rotate( &Left, Top, -0.5 * Pi );

  //distinguish between keys
  switch( Key )
  {
    //escape
    case 27:
      //toggle console
      Console = !Console;
      //display message
      if( Console )
        ConsoleInsertLine( "[escape] => console on" );
      else
        ConsoleInsertLine( "[escape] => console off" );
      break;
    //enter
    case '\r':
    case '\n':
      if( Console ) //console on
      {
        //get last line of console
        GetConsoleLine( Line, sizeof( Line ) );
        //display enter
        ConsolePutc( '\n' );
        //process Command
        ProcessCommand( Line );
      }
      else //console off
      {
        //go to start position
        SetVector( &Pos, CenterX, CenterY, CenterZ );
        SetVector( &Front, 0.0, 1.0, 0.0 );
      }
      break;
    //backspace
    case '\b':
      if( Console ) //console on
        ConsolePutc( '\b' );
      break;
    //space
    case ' ':  
      if( Console ) //console on
        ConsolePutc( ' ' );
      else //console off
        //restore horizontal view
        Front.Z = 0.0;
      break;
    //delete
    case 127:
      if( !Console ) //console off
        //move right
        LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Left );
      break;
    //normal character
    default:
      if( Console && Key > ' ' )
        ConsolePutc( Key );
  } //switch( Key )

  //repair position and front direction
  RepairPosFront( );
}

//glut special keyboard callback
static void GlutSpecial( int Key, int X, int Y )
{
  //local variables
  tVector Dir, Top, Left;

  //keep compiler happy
  X = 0;
  Y = 0;

  //get direction over ground
  Dir = Front;
  Dir.Z = 0;
  Normalize( &Dir );
  //get top and left vectors
  SetVector( &Top, 0.0, 0.0, 1.0 );
  Left = Dir;
  Rotate( &Left, Top, -0.5 * Pi );

  //distinguish between keys
  switch( Key )
  {
    //F4
    case GLUT_KEY_F4:
      //end program
      RequestEnd = 1;
      break;
    //left arrow
    case GLUT_KEY_LEFT:
      if( Console ) //console on
        //move console left
        ConsoleMoveLeft( );
      else //console off
        //turn left
        Rotate( &Front, Top, - TurnStep );
      break;
    //right arrow
    case GLUT_KEY_RIGHT:
      if( Console ) //console on
        //move console right
        ConsoleMoveRight( );
      else //console off
        //turn right
        Rotate( &Front, Top, TurnStep );
      break;
    //up arrow
    case GLUT_KEY_UP:
      if( Console ) //console on
        //move console up
        ConsoleMoveUp( );
      else //console off
        //move forwards
        LinearCombine( &Pos, 1.0f, Pos, MoveStep, Dir );
      break;
    //down arrow
    case GLUT_KEY_DOWN:
      if( Console ) //console on
        //move console down
        ConsoleMoveDown( );
      else //console off
        //move backwards
        LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Dir );
      break;
    //home
    case GLUT_KEY_HOME:
      if( ! Console ) //console off
        //look up
        Rotate( &Front, Left, TurnStep );
      break;
    //end
    case GLUT_KEY_END:
      if( ! Console ) //console off
        //look down
        Rotate( &Front, Left, - TurnStep );
      break;
    //page up
    case GLUT_KEY_PAGE_UP:
      if( Console ) //console on
        //move console up a page
        ConsoleMovePageUp( );
      else //console off
        //move up
        LinearCombine( &Pos, 1.0f, Pos, MoveStep, Top );
      break;
    //page down
    case GLUT_KEY_PAGE_DOWN:
      if( Console ) //console on
        //move console down a page
        ConsoleMovePageDown( );
      else //console off
        //move down
        LinearCombine( &Pos, 1.0f, Pos, - MoveStep, Top );
      break;
    //insert
    case GLUT_KEY_INSERT:
      if( !Console ) //console off
        //move left
        LinearCombine( &Pos, 1.0f, Pos, MoveStep, Left );
      break;
  } //switch( Key )

  //repair position and front direction
  RepairPosFront( );
}

//cleanup procedure
static void AtExit( )
{
  //cleanupt stream recpetion
  StreamRecvStop( );

  //unload cam-points
  free( pCampts );

  //cleanup script buffer
  ScriptBufferCleanup( );

  //close logfile
  LogStop( );
}

//main program
int main( int ParamCnt, char * * Params )
{
  //local variables
  long I, X, Y, B, F;
  char Text[256];

  //global initialisations
  srand( time( NULL ) );

  //initialize script buffer
  ScriptBufferSet( "scriptload TroiaPreview.ini" );

  //read parameters
  ScreenX = 640;
  ScreenY = 480;
  ScreenBits = 16;
  ScreenFreq = 60;
  FullScreen = 1;
  for( I = 0; I < ParamCnt; I++ )
  {
    //script=<file>: load script file
    if( strncasecmp( Params[I], "script=", 7 ) == 0 )
    {
      snprintf( Text, sizeof( Text ), "scriptload %s", Params[I] + 7 );
      ScriptBufferSet( Text );
    }
    //log=<file>: generate log file
    if( strncasecmp( Params[I], "log=", 4 ) == 0 )
    {
      LogStart( Params[I] + 4 );
    }
    //window=<width>x<height>: use window mode
    if( strncasecmp( Params[I], "window=", 7 ) == 0 )
    {
      if( sscanf( Params[I] + 7, "%ldx%ld", &X, &Y ) == 2 )
      {
        ScreenX = LongMinMax( X, 640, 6400 );
        ScreenY = LongMinMax( Y, 400, 4000 );
        FullScreen = 0;
      }
    }
    //fullscreen=<width>x<height>:<bits>@<freqency>: use fullscreen mode
    if( strncasecmp( Params[I], "fullscreen=", 11 ) == 0 )
    {
      if( sscanf( Params[I] + 11, "%ldx%ld:%ld@%ld", &X, &Y, &B, &F ) == 4 )
      {
        ScreenX = LongMinMax( X, 640, 6400 );
        ScreenY = LongMinMax( Y, 400, 4000 );
        ScreenBits = LongMinMax( B, 1, 256 );
        ScreenFreq = LongMinMax( F, 10, 1000 );
        FullScreen = 1;
      }
    }
  } //for( I ...
  
  //initialize glut
  glutInit( &ParamCnt, Params );
  glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
  glutInitWindowPosition( 0, 0 );
  glutInitWindowSize( ScreenX, ScreenY );
  if( FullScreen )
  {
    sprintf( Text, "%ldx%ld:%ld@%ld", ScreenX, ScreenY, ScreenBits, ScreenFreq );
    glutGameModeString( Text );
    glutEnterGameMode( );
  }
  else
  {
    glutCreateWindow( "TroiaPreview" );
  }
  glutDisplayFunc( GlutDisplay );
  glutIdleFunc( GlutIdle );
  glutKeyboardFunc( GlutKeyboard );
  glutSpecialFunc( GlutSpecial );

  //OpenGl settings
  glDepthFunc( GL_LEQUAL );
  glEnable( GL_DEPTH_TEST );
  glEnable( GL_AUTO_NORMAL );
  glEnable( GL_NORMALIZE );
  glPixelStorei( GL_UNPACK_ALIGNMENT, 4 );
  glViewport( 0, 0, ScreenX, ScreenY );
  glLineWidth( 1.5 ); //set line width
  glPointSize( 2.0 ); //set point size

  //initialize cam-points
  CamptsCnt = 0; //no cam-points
  pCampts = (tCampt *)malloc( 0 );
  CamptSpeed = 0.0; //cam-point-speed
  CamptRefTime = 0.0; //reference-time for cam-point-movement
  CamMoving = 0; //if cam is not moving

  //start-position
  SetVector( &Pos, CenterX, CenterY, CenterZ );
  SetVector( &Front, 0.0, 1.0, 0.0 );

  //initialise console
  ConsoleInit( );
  ConsoleInsertLine( "OpenGL information" );
  strcpy( Text, "  GL_VENDOR: " );
  strncat( Text, (char *)glGetString( GL_VENDOR ), sizeof( Text ) - strlen( Text ) - 1 );
  ConsoleInsertLine( Text );
  strcpy( Text, "  GL_RENDERER: " );
  strncat( Text, (char *)glGetString( GL_RENDERER ), sizeof( Text ) - strlen( Text ) - 1 );
  ConsoleInsertLine( Text );
  strcpy( Text, "  GL_VERSION: " );
  strncat( Text, (char *)glGetString( GL_VERSION ), sizeof( Text ) - strlen( Text ) - 1 );
  ConsoleInsertLine( Text );
  ConsoleInsertLine( "TroiaPreview" );
  ConsoleInsertLine( "  version 0.4 date 2004-10-23" );
  ConsoleInsertLine( "  Copyright (C) 2003-2004 1stein <1stein@schuermans.info>" );
  ConsoleInsertLine( "  Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html" );
  ConsoleInsertLine( "  some important keys:" );
  ConsoleInsertLine( "    F4: exit program" );
  ConsoleInsertLine( "    escape: toggle console" );
  ConsoleInsertLine( "  when console is off:" );
  ConsoleInsertLine( "    arrow-keys: move around" );
  ConsoleInsertLine( "  when console is on:" );
  ConsoleInsertLine( "    arrow-keys: scroll console" );
  ConsoleInsertLine( "  type \"help\" for commands" );

  //set time to 0
  TimeSet( 0.0 );

  //initialize frame with random data
  FrameRandomize( 0x00, 0xFF );

  //initialize
  TickerInit( );

  //register cleanup procedure
  atexit( AtExit );

  //main loop
  glutMainLoop( );

  //end
  return 0;
}
